/* $Id: et6000.c,v 1.13 1999/06/27 16:01:15 marcus Exp $
******************************************************************************

   Tseng Labs ET6000 chipset driver

   Copyright (C) 1997 Jim Ursetto	[jim.ursetto@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
 
	This code was derived from the following sources of information:

	[1] Prerelease VGADOC5 (Tseng.txt/vgaregs.txt) by Finn Thoegersen
	[2] IBM VGA driver by Jason McMullan
	Copyright 1990,91 by Thomas
	Roell, Dinkelscherben, Germany
	[4] S3 86c765 (Trio64V+) driver by Jon Taylor

******************************************************************************
*/

#include <kgi/maintainers.h>

#define	MAINTAINER	Jim_Ursetto
#define DRIVER_NAME	"Tseng Labs ET6000 chipset driver"
#define DRIVER_REV	"$Revision: 1.13 $"

#define DEBUG_LEVEL	0

#include <kgi/system.h>
#include <kgi/module.h>

#include <IBM/vga.h>
#include <Tseng/et6000.h>
#include <Tseng/et6000_fb.h>
#include <Tseng/et6000_io.h>
#include <Tseng/et6000_text16.h>

/* !! get rid of this */
/*#include <IBM/vga_io.inc>*/

#undef 	CURSOR_HW	/* Hardware cursor */

#define TEXT_LEN 0x20000

#define	POINTER_FONT	/* Font-hack mouse cursor in textmodes */
#define SLOW_POINTER	/* #define this if your pointer flickers */

static struct kgi_chipset chipinfo =
{
	"Tseng Labs",		/* Manufacturer		*/
	"ET6000",		/* Model		*/
	GM_TEXT16 | GM_1BIT |	/* Available modes	*/
	GM_4BIT | GM_8BIT | GM_15BIT | GM_16BIT | GM_24BIT | GM_32BIT,
	0,			/* Flags		*/
	{ 2048, 2048 },		/* Maximum resolution	*/
	NULL,			/* Private data hook	*/
	0x0,			/* addressable RAM (autodetect)         */
	{ 0000000, 135000000 }	/* DCLK limits				*/
};

static struct mem_region et6000_linear_mem =
{
	MEM_NULL,
	0x0,           /* linear framebuffer (autodetect) */
	0x0,           /* mem size (autodetect) */
	-1,            /* Decode all bits */
	"Tseng Labs ET6000 linear framebuffer"
};

static struct mem_region et6000_banked_mem =
{
	MEM_NULL,
	0xA0000,
	0x10000,
	-1,
	"Tseng Labs ET6000 banked framebuffer"
};

static struct io_region et6000_io =
{
	IO_NULL,
	0x3C0, /* VGA IO Region Start */
	0x20, /* 32 register (0x3C0-0x3DF) */
	0x3FF, /* Only 10 address lines are decoded.. */
	"Tseng Labs ET6000 I/O region"
};    

struct io_region et6000_pci_io = 
{
	IO_NULL,
	0,
	0x87, /* Number of PCI registers */
	-1,
	"Tseng Labs ET6000 PCI I/O region"
};

/* Unused.  Accelerator registers are mapped as part of 4 MB framebuffer. */

#if 0
static struct mem_region et6000_accel =
{
	MEM_NULL,
	0,	    /* base ptr */
	0xFF,     /* size */
	-1,       /* Decode all bits */
	"Tseng Labs ET6000 Accelerator region"
};
#endif

static struct kgi_mmio_region et6000_linear_mmio;

struct chipset_private
{
	kgiu32 linear_mem_size;
};

#define KGI_POKEBITS(VAL,MASK,REG) REG = ((REG & ~MASK) | (MASK & VAL))

#define GET_GRAPH_PRIVATE \
		struct graphics_private *gpriv = KGIM_PRIVATE(dpy, graphics)

#define GET_PRIVATE \
		struct chipset_private *priv = KGIM_PRIVATE(dpy, chipset)

#define GET_PCI \
		struct kgim_options_pci *pci = KGIM_OPTIONS(dpy, pci)

#define PCI_IO (et6000_pci_io.baseptr)
#define PCI_IO_REGION (et6000_pci_io)
#define VGA_IO (et6000_io.baseptr)
#define VGA_IO_REGION (et6000_io)
#define LINEAR_MEM_REGION (et6000_linear_mem)
#define LINEAR_MEM_SIZE (priv->linear_mem_size)
#define LINEAR_MEM (et6000_linear_mem.baseptr)
#define LINEAR_MMIO (et6000_linear_mmio)

/* This table holds the default values for the VGA registers.  A copy	*/
/* of it will be filled in as appropriate during mode programming in	*/
/* kgim_chipset_set_mode().						*/

static struct statetable default_state =
{
	0x23,							/* MISC	*/
	{
		0x03, 0x00, 0x00, 0x00, 0x00			/* SEQ	*/
	},
	{
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* CRT	*/
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x1x	*/
		0x00
	},
	{
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* GRC	*/
		0xFF
	},
	{
		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,	/* ATC	*/
		0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
		0x00, 0x00, 0x0F, 0x00, 0x00
	},
		0x00                                                    /* VidCtrl1 */
};

struct statetable reset_state;
kgi_modereq default_kgi_modereq;
struct kgi_mode default_kgi_mode;

static kgiu32 calc_mem(void);

static int verify_et6000(struct kgi_display *dpy)
{
	kgiu16 status;
	kgiu8 result8;
	kgiu32  result32;
	kgiu32 et6kbase;
	kgiu8 bus, device_fn;
	GET_PCI;

	/* Check if address parameters were passed; otherwise, probe.  Only the
	 * first card will be found. */
	if (pci->dev == PCICFG_NULL)
	{
#ifdef __linux__
		NOTICE("Probing the PCI bus.");
		status = pcibios_find_device(PCI_VENDOR_ID_TSENG, 
			 PCI_DEVICE_ID_TSENG_ET6000, 0, &bus, &device_fn);
	
		if (status != PCIBIOS_SUCCESSFUL)
		{
			NOTICE("PCI probe did not find an ET6000.");
			return -E(CHIPSET, UNKNOWN);
		}
		/* We found an ET6000.  Convert the autoprobed address values into
		 * GGI-compatible ones. */
		pci->dev = PCICFG_VADDR(bus, PCI_SLOT(device_fn), PCI_FUNC(device_fn));
#else
		/* Hopefully a probe wrapper will be added to GGI. */
		NOTICE("PCI probing is not yet supported on non-Linux systems.");
		return -E(CHIPSET, UNKNOWN);
#endif
	}
	else
	{
		NOTICE("Using user-supplied PCI address.");
		if ((pcicfg_in16(pci->dev + PCI_VENDOR_ID) != PCI_VENDOR_ID_TSENG)
			|| ((pcicfg_in16(pci->dev + PCI_DEVICE_ID)) != PCI_DEVICE_ID_TSENG_ET6000))
		{
			NOTICE("An ET6000 was not found at the specified address.");
			return -E(CHIPSET, UNKNOWN);                    
		}
	}

/*  !! LOOKHERE Make sure that misc register is in a sane state. */
/*	outbMISC(inb(MISCr));*/	/* Adjust IO base       */
    
	result8 = pcicfg_in8(pci->dev + PCI_REVISION_ID);

	/* I want this to be printk with KERN_INFO priority, but it's KERN_ERR */
	if (result8 < 70)
		NOTICE("Found ET6000 chipset, revision %d.", result8);
	else {
		NOTICE("Found ET6100 (rev. %d)--not fully supported yet.", result8);
		NOTICE("Please report this message to " MAINTAINER ".");
	}
	
	result32 = pcicfg_in32(pci->dev + PCI_BASE_ADDRESS_0);
	result32 &= 0xff000000;
	DEBUG1("Linear framebuffer is at 0x%08x.\n", result32);
	et6kbase = pcicfg_in32(pci->dev + PCI_BASE_ADDRESS_1);
	et6kbase &= ~0xff;
	DEBUG1("PCI I/O is mapped at port 0x%04x.\n", et6kbase);
	
	/* Now that we've got the I/O address of PCI space, we'll use that
	 * instead of the clunky GGI wrappers.  The primary reason for this is
	 * that the BIOS functions don't seem to work right all of the time.
	 * Whether or not this is only true with my particular card is unknown,
	 * but using I/O is safe in any case.  The functions PCI_in8 and
	 * PCI_out8 are this driver's own PCI wrappers, and could easily be
	 * rewritten to use BIOS calls.
	 */
	
	PCI_IO_REGION.base = et6kbase;
	LINEAR_MEM_REGION.base = result32;

	chipinfo.ramsize = calc_mem();
	if (!chipinfo.ramsize)
		return -E(CHIPSET, UNKNOWN);

	/* We _must_ mmap 4096 KB (for accel), and set the true size later. */
	et6000_linear_mem.size = 4096 KB;
	et6000_linear_mmio.next	= NULL;
	et6000_linear_mmio.name	= "ET6000 Video RAM";
	et6000_linear_mmio.type	= MMAP_TYPE_MMIO | MMAP_PUBLIC | MMAP_FRAMEBUFFER;
	et6000_linear_mmio.prot	= MMIO_RDWR | MMIO_SHARED;
	et6000_linear_mmio.psize    = chipinfo.ramsize;
	et6000_linear_mmio.size     = chipinfo.ramsize;
	et6000_linear_mmio.pstart   = LINEAR_MEM_REGION.base;
	return EOK;
}

/* This memory-detection code is from the XFree86 driver. */
#define VIDMEM ((unsigned int *)et6000_banked_mem.baseptr)
#define SEGSIZE 64 /* KB */

static void Tseng_set_segment(int seg)
{
  int seg1, seg2;
  seg1 = seg & 0x0F;
  seg2 = (seg & 0x30) >> 4;
  io_out8(seg2 | (seg2 << 4), 0x3CB);
  io_out8(seg1 | (seg1 << 4), 0x3CD);
}

static int et6000_check_videoram(int ram)
{
  kgiu8 oldSegSel1, oldSegSel2, oldGR5, oldGR6, oldSEQ2, oldSEQ4;
  int segment, i;
  int real_ram = 0;

  int fooled = 0;
  int save_vidmem;

  /* Temporarily commandeer the VGA memory area for banked use. */
  if (mem_check_region(&et6000_banked_mem))
	return 0;

  mem_claim_region(&et6000_banked_mem);
  
 /*
  * We need to set the VGA controller in VGA graphics mode, or else we won't
  * be able to access the full 4MB memory range. First, we save the
  * registers we modify, of course.
  */
  oldSegSel1 = io_in8(0x3CD);
  oldSegSel2 = io_in8(0x3CB);
  io_out8(5, 0x3CE); oldGR5 = io_in8(0x3CF);
  io_out8(6, 0x3CE); oldGR6 = io_in8(0x3CF);
  io_out8(2, 0x3C4); oldSEQ2 = io_in8(0x3C5);
  io_out8(4, 0x3C4); oldSEQ4 = io_in8(0x3C5);

  /* set graphics mode */  
  io_out8(6, 0x3CE); io_out8(5, 0x3CF);
  io_out8(5, 0x3CE); io_out8(0x40, 0x3CF);
  io_out8(2, 0x3C4); io_out8(0x0f, 0x3C5);
  io_out8(4, 0x3C4); io_out8(0x0e, 0x3C5);

 /*
  * count down from presumed amount of memory in SEGSIZE steps, and
  * look at each segment for real RAM.
  */

  for (segment = (ram / SEGSIZE) - 1; segment >= 0; segment--)
  {
      Tseng_set_segment(segment);

      /* save contents of memory probing location */
      save_vidmem = *(VIDMEM);

      /* test with pattern */
      *VIDMEM = 0xAAAA5555;
      if (*VIDMEM != 0xAAAA5555) {
          *VIDMEM = save_vidmem;
          continue;
      }
      
      /* test with inverted pattern */
      *VIDMEM = 0x5555AAAA;
      if (*VIDMEM != 0x5555AAAA) {
          *VIDMEM = save_vidmem;
          continue;
      }
      
      /* check if we aren't fooled by address wrapping (mirroring) */
      fooled = 0;
      for (i = segment-1; i >=0; i--) {
          Tseng_set_segment(i);
          if (*VIDMEM == 0x5555AAAA) {
               fooled = 1;
               break;
          }
      }
      if (!fooled) {
          real_ram = (segment+1) * SEGSIZE;
          break;
      }
      /* restore old contents again */
      *VIDMEM = save_vidmem;
  }

  /* restore original register contents */  
  io_out8(oldSegSel1, 0x3CD);
  io_out8(oldSegSel2, 0x3CB);
  io_out8(5, 0x3CE); io_out8(oldGR5, 0x3CF);
  io_out8(6, 0x3CE); io_out8(oldGR6, 0x3CF);
  io_out8(2, 0x3C4); io_out8(oldSEQ2, 0x3C5);
  io_out8(4, 0x3C4); io_out8(oldSEQ4, 0x3C5);

  mem_free_region(&et6000_banked_mem);

  return real_ram;
}
#undef VIDMEM
#undef SEGSIZE

static kgiu32 calc_mem(void)
{
	int type;
	kgiu8 size;
	kgiu32 mem_size = 0;
	kgiu32 real_size = 0;	
	enable_ext();
	type = io_in8(MISCw) & 0x03;
	disable_ext();

	switch(type) {
		case 0x03:
			size = PCI_in8(NULL, PCI_TSTSETUP);
			mem_size = ((size & PCI47_RFN) + 1) << 8; /*  * 32K * 8 banks */

			if (PCI_in8(NULL, PCI_MEMCONF) & PCI45_MDRAMDBL)
				 mem_size <<= 1;
			
			NOTICE("%d KB MDRAM predicted.", mem_size);

			/* Check the real memory size.  I've received reports of < 4 MB
			 * being reported as 4, and there's the 2.25->2.5 thingy the X
			 * driver mentions.
			 */
			if (!(real_size=et6000_check_videoram(mem_size))) {
				NOTICE("Physical memory check failed.  Using predicted.\n");
			} else mem_size = real_size;
		
			NOTICE("%d KB MDRAM detected.", mem_size);
			break;

		case 0x00:
			size = PCI_in8(NULL, PCI_MEMCONF);
			mem_size = 1024 << (size & PCI45_DRAMSIZE);
			NOTICE("%d KB DRAM detected.", mem_size);
			break;
		default:
			ERROR("Video RAM detection failed!\n");
    }
	return mem_size << 10;
}
 
/* GGI interface */
	 
/* Mode specific frame buffers */

static void chipset_setup_dpy(struct kgi_display *dpy, struct kgi_mode *mode)
{
	switch (mode->request.graphtype) {
		case KGIGT_TEXT16:
			dpy->set_font       = text16_set_font;
			dpy->set_origin     = text16_set_origin;
			dpy->set_split      = text16_set_split;
			dpy->set_frame      = text16_set_frame;
			dpy->show_cursor    = text16_show_cursor;
			dpy->hide_cursor    = text16_hide_cursor;
			dpy->undo_cursor    = text16_undo_cursor;
			dpy->show_pointer   = text16_show_pointer;
			dpy->hide_pointer   = text16_hide_pointer;
			dpy->undo_pointer   = text16_undo_pointer;
			dpy->fb.dpy         = dpy;
			dpy->fb.size.x      = mode->request.virt.x;
			dpy->fb.size.y      = mode->request.virt.y;
			dpy->fb.bpp         = MAP_16_BPP;
	        dpy->fb.pbuf        = LINEAR_MEM;
			dpy->fb.zbuf        = NULL;
			dpy->fb.ops         = &text16_color_io;
	    
			/* This overwrote all the KGIM_PRIVATE data before. */
			KGIM_PRIVATE(dpy, driver) = text16_data_realloc(KGIM_PRIVATE(dpy, driver));
			if (KGIM_PRIVATE(dpy, driver) == NULL)
				INTERNAL_ERROR;
		
			break;
		default:
			dpy->set_font		= NULL;
			dpy->set_origin		= et6000_fb_set_origin;
			dpy->set_split		= et6000_fb_set_split;
			dpy->set_frame		= et6000_fb_set_frame;
			dpy->show_cursor	= NULL;
			dpy->hide_cursor	= NULL;
			dpy->undo_cursor	= NULL;
			dpy->show_pointer	= NULL;
			dpy->hide_pointer	= NULL;
			dpy->undo_pointer	= NULL;
			dpy->fb.dpy			= dpy;
			dpy->fb.size.x		= 0;
			dpy->fb.size.y		= 0;
			dpy->fb.bpp			= 0;
			dpy->fb.pbuf		= NULL;
			dpy->fb.zbuf		= NULL;
			dpy->fb.ops			= NULL;
/*/ AJA:		kgim_graphics_set_mode(dpy,mode); */
			kgim_accel_set_mode(dpy,mode);
	}
	memcpy(&(dpy->mode), mode, sizeof(struct kgi_mode));
}

/* Takes a mode and a bits-per-pixel value */
static void chipset_setup_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	int bpp=0;
	GET_PRIVATE;
    
	/* By default, setup for graphics modes */

	switch (mode->request.graphtype) {
		case KGIGT_TEXT16:
			mode->fb	= LINEAR_MEM;
			mode->fb_size	= TEXT_LEN;
			mode->fsize	= mode->request.virt.x * mode->request.virt.y * 2;
			mode->mem	= &text16_color_mem;
			mode->io	= &text16_color_io;
			mode->mmio  = &et6000_linear_mmio;
			return; 	/* Done setting up */
	    
		case KGIGT_1BIT:
			bpp = 1;
			break;
		case KGIGT_4BIT:
			bpp = 4;
			break;
		case KGIGT_8BIT:
			bpp = 8;
			break;
		case KGIGT_15BIT:
		case KGIGT_16BIT:
	    	bpp = 16;
	    	break;
		case KGIGT_24BIT:
		    bpp = 24;
	    	break;
		case KGIGT_32BIT:
		    bpp = 32;
	    	break;
		default:
			INTERNAL_ERROR;
			return;
	}

	mode->fb = LINEAR_MEM;
	mode->fb_size = LINEAR_MEM_SIZE;
	mode->fsize	= (mode->request.virt.x * 
	               mode->request.virt.y * 8) / bpp;

	DEBUG2("requested x: %d y: %d vx: %d vy: %d\n", 
	        mode->request.visible.x, mode->request.visible.y,
	        mode->request.virt.x, mode->request.virt.y);
	
	mode->mem	= NULL;
	mode->io	= NULL;
	mode->mmio  = &et6000_linear_mmio;
}

/* Does not save extended registers!!! */
static int chipset_get_reset_mode(struct kgi_display *dpy)
{
	kgi_graphtype gmode = KGIGT_TEXT16;
	int magx = 1, magy = 1;
	int DCLKdiv = 8;
	int dotclock;
	struct kgi_timing tx,ty;
	save_state(&reset_state);

	if (reset_state.ATC[0x10] & AR10_GRAPHICS) {
		switch (reset_state.GRC[0x05] & GR5_WRITEMODE3) {
			case GR5_WRITEMODE0:
			    gmode = KGIGT_8BIT; /* 8 bit mode */
				break;
			case GR5_WRITEMODE1:
				gmode = KGIGT_1BIT; /* 1 bit mode */
				break;
			case GR5_WRITEMODE3:
				gmode = KGIGT_4BIT; /* 4 bit mode */
				break;
			case GR5_WRITEMODE2:
				break;
		}
		magy = (reset_state.CRT[0x09] & CR09_CHARHEIGHT_MASK) + 1;
	} else { /* Text mode */
		gmode = KGIGT_TEXT16;
		DCLKdiv = (reset_state.SEQ[0x01] & SR01_8DOT_CHCLK) ? 8 : 9;
	}
    
	/* Get horizontal timing */
	tx.width	= (reset_state.CRT[0x01]+1) * DCLKdiv;
	tx.blankstart	= reset_state.CRT[0x02] * DCLKdiv;
	tx.syncstart	= reset_state.CRT[0x04] * DCLKdiv;
	tx.syncend	= ((reset_state.CRT[0x05]) & CR05_HSE_MASK) * DCLKdiv;
	while (tx.syncend < tx.syncstart)
		tx.syncend += 0x20 * DCLKdiv;
	tx.blankend	= ((reset_state.CRT[0x03] & CR03_HBE_MASK) |
			   ((reset_state.CRT[0x05] & CR05_HBLANKEND_B5) ? 0x20 : 0) |
			   (reset_state.CRT[0x01] & 0xC0)) * DCLKdiv;
	tx.total	= (reset_state.CRT[0x00]+5) * DCLKdiv;
	tx.polarity	= (reset_state.MISC & MISC_NEG_HSYNC) != 0;

	/* Get vertical timing */
	ty.width	= 1 + (reset_state.CRT[0x12] | 
			       ((reset_state.CRT[0x07] & CR07_VDISPLAYEND_B8) ? 0x100 : 0) |
			       ((reset_state.CRT[0x07] & CR07_VDISPLAYEND_B9) ? 0x200 : 0));
	ty.blankstart	= 1 + (reset_state.CRT[0x15] |
			       ((reset_state.CRT[0x07] & CR07_VBLANKSTART_B8) ? 0x100 : 0) |
			       ((reset_state.CRT[0x09] & CR09_VBLANKSTART_B9) ? 0x200 : 0));
	ty.syncstart	= (reset_state.CRT[0x10] |                
			   ((reset_state.CRT[0x07] & CR07_VSYNCSTART_B8) ? 0x100 : 0) |
			   ((reset_state.CRT[0x07] & CR07_VSYNCSTART_B9) ? 0x200 : 0));
	ty.syncend	= (reset_state.CRT[0x11] & CR11_VSYNCEND_MASK) |
                (reset_state.CRT[0x12] & ~CR11_VSYNCEND_MASK);
	while (ty.syncend < ty.syncstart)
		ty.syncend += 0x10;
	ty.blankend	= (reset_state.CRT[0x16] | ((ty.width - 1) & ~0xFF)) + 1;
	ty.total = 2 + (reset_state.CRT[0x06] |
			((reset_state.CRT[0x07] & CR07_VTOTAL_B8) ? 0x100 : 0) |
			((reset_state.CRT[0x07] & CR07_VTOTAL_B9) ? 0x200 : 0));
	ty.polarity = (reset_state.MISC & MISC_NEG_VSYNC) != 0;
	
/*/ AJA: dotclock = (getCLK() == MISC_25MHZ_CLK) ? 25175000 : 28322000; */
	dotclock = (1) ? 25175000 : 28322000;
   
	/* Set all the mode info that we've just gathered */
	default_kgi_modereq.frames     = 1;
	default_kgi_modereq.dpp.x      = DCLKdiv;
	default_kgi_modereq.dpp.y      = (reset_state.CRT[0x09] & CR09_CHARHEIGHT_MASK) + 1;
	default_kgi_modereq.visible.x  = tx.width / default_kgi_modereq.dpp.x;
	default_kgi_modereq.visible.y  = ty.width / default_kgi_modereq.dpp.y;
	default_kgi_modereq.virt.x     = tx.width / default_kgi_modereq.dpp.x;
	default_kgi_modereq.virt.y     = ty.width / default_kgi_modereq.dpp.y;
	default_kgi_modereq.size.x     = 0;
	default_kgi_modereq.size.y     = 0;
	default_kgi_modereq.graphtype  = gmode;
	if (dpy->defmode == NULL)
		dpy->defmode = &default_kgi_modereq;
	else
		(*dpy->defmode) = default_kgi_modereq;

	dpy->mode.request	= default_kgi_modereq;
	dpy->mode.x		= tx;
	dpy->mode.y		= ty;
	dpy->mode.magnify.x	= magx;
	dpy->mode.magnify.y	= magy;
	dpy->mode.dclk		= dotclock;
	dpy->mode.clk.mul	= 1;
	dpy->mode.clk.div	= 1;
	dpy->mode.lclk.mul	= 1;
	dpy->mode.lclk.div	= 1;
	dpy->mode.chipset	= 0;
	
	default_kgi_mode = dpy->mode;

	chipset_setup_mode(dpy, &default_kgi_mode);
	chipset_setup_dpy(dpy, &(default_kgi_mode));

	return EOK;
}
	 
static int chipset_init(struct kgi_display *dpy)
{
	int err;
	kgiu8 result8;
	GET_PRIVATE;
    
	NOTICE(DRIVER_NAME " " DRIVER_REV);

	if (verify_et6000(dpy) == EOK) {
#ifdef YES_WE_REALLY_WANT_TO_CHECK_IO_REGIONS
		if ((err = io_check_region(&VGA_IO_REGION)))
			return err;
		if ((err = io_check_region(&PCI_IO_REGION)))
			return err;
#endif
		if ((err = mem_check_region(&LINEAR_MEM_REGION)))
			return err;

	    mem_claim_region(&LINEAR_MEM_REGION);
 	    io_claim_region(&VGA_IO_REGION);
	    io_claim_region(&PCI_IO_REGION);
		
		/* Only 4,088 KB can be used due to overlap of mapped
		 * accel registers and framebuffer. */
		/* We put lfb size into private data, so the graphics module can
		 * subtract what it needs for scratchpad. */
		/* Warning: LINEAR_MEM_SIZE reflects actual memory present.
		 * et6000_linear_mem.size is always 4096KB, the mmapped amount. */
		LINEAR_MEM_SIZE = chipinfo.ramsize > 4088 KB ? 4088 KB
		                : chipinfo.ramsize;

	    /* Now that we've commandeered PCI I/O region, we can stop using BIOS calls. */
		result8 = PCI_in8(dpy, PCI_MMCONTROL);
		PCI_out8(dpy, result8 | PCI40_LINEARMAP | PCI40_MEMRELOC
		         | PCI40_MEMMAPREG, PCI_MMCONTROL);
		text16_data_init();

		return chipset_get_reset_mode(dpy);
	}

	return -E(CHIPSET,FAILED);
}
	 
static void chipset_reset(struct kgi_display *dpy)
{
	restore_state(&reset_state);
	if (dpy) {
		dpy->mode = default_kgi_mode;
		(*dpy->defmode) = default_kgi_modereq;
	}
}

static void chipset_done(void)
{
/*	struct kgi_display *dpy = NULL;
	chipset_reset(NULL);
        
	PCI_out8(dpy, PCI_in8(dpy, PCI_MMCONTROL) &~ PCI40_LINEARMAP
			 &~ PCI40_MEMRELOC &~ PCI40_MEMMAPREG, PCI_MMCONTROL);

	io_free_region(&VGA_IO_REGION);
	io_free_region(&PCI_IO_REGION); */
	mem_free_region(&et6000_linear_mem);
}
	 
/* I'm having problems at high refresh with screen distortion; I increased horizontal
 * total _in my own monitor driver_ for now. */
static void adjust_timing(struct kgi_mode *mode)
{
	register int foo = (mode->request.graphtype == KGIGT_TEXT16)
		? mode->request.dpp.x : 8;

#define ADJUST(X) mode->x.X -= (mode->x.X % foo)

	ADJUST(width);
	ADJUST(blankstart);
	ADJUST(syncstart);
	ADJUST(syncend);
	ADJUST(blankend);
	if (mode->x.total % foo)
		mode->x.total += foo - (mode->x.total % foo);

	if (((mode->x.syncend / foo) & 0x3F) == ((mode->x.syncstart / foo) & 0x3F))
		mode->x.syncend -= foo;
	if (((mode->x.blankend / foo) & 0x7F) == ((mode->x.blankstart / foo) & 0x7F))
		mode->x.blankend -= foo;

	if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F))
		mode->y.syncend--;
	if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF))
		mode->y.blankend--;

#undef ADJUST
}

/* Consider implementing roundmode, as VGA driver does. */

static int chipset_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
                              enum kgi_tm_cmd cmd)
{
	int bpp = 0;
	int textx, texty;
	GET_PRIVATE;
    
	textx = mode->request.dpp.x;
	texty = mode->request.dpp.y;

	switch (mode->request.graphtype) {
		case KGIGT_8BIT:
			bpp = 8;	/* 1 Bpp */
			break;
		case KGIGT_15BIT:
		case KGIGT_16BIT:
	    	bpp = 16;
		    break;
		case KGIGT_24BIT:
		    bpp = 24;
	    	break;
		case KGIGT_32BIT:
		    bpp = 32;
	    	break;
		case KGIGT_1BIT:
			bpp = 1;  /* LOOKHERE -- this was changed from 4 */
			break;
		case KGIGT_4BIT:
			bpp = 4;	/* 0.5 Bpp */
			DEBUG0("4 bpp mode is not supported yet.");
			return -E(CHIPSET, FAILED);
		case KGIGT_TEXT16:
			break;		/* Not used in textmode */
		default:
			DEBUG0("graphtype %i not supported", mode->request.graphtype);
			return -E(CHIPSET,FAILED);
	}

#define XVIRTUAL (mode->request.virt.x)
#define YVIRTUAL (mode->request.virt.y)
#define XVISIBLE (mode->request.visible.x)
#define YVISIBLE (mode->request.visible.y)

	switch (cmd) {
		case CMD_PROPOSE:
			if (mode->request.graphtype == KGIGT_TEXT16) {

				mode->magnify.x = mode->magnify.y = 1;
				if ((textx < 8) || (9 < textx)) {
					DEBUG0("Invalid char width (%i)", textx);
					return -EINVAL;
				}

/*				mode->request.visible.x -=
				(mode->request.visible.x % textx);
				mode->request.visible.y -=
				(mode->request.visible.y % texty);
*/
				if (XVIRTUAL == 0 && YVIRTUAL == 0)
					return -EINVAL;

				if (XVIRTUAL == 0) 
					XVIRTUAL = TEXT_LEN / (YVIRTUAL * 2);
			
				if (YVIRTUAL == 0)
					YVIRTUAL = TEXT_LEN / (XVIRTUAL * 2);

				if ((2 * XVIRTUAL * YVIRTUAL) > TEXT_LEN) {
					DEBUG1("Not enough text memory");
					return -ENOMEM;
				}
				mode->x.width = mode->request.visible.x * textx;
				mode->y.width = mode->request.visible.y * texty;

			} else {	/* mode->request.graphtype == KGIGT_TEXT16 */

				if (XVIRTUAL == 0 && YVIRTUAL == 0)
					return -EINVAL;

				if (XVIRTUAL == 0)
					XVIRTUAL =  LINEAR_MEM_SIZE / YVIRTUAL;

				if (YVIRTUAL == 0)
					YVIRTUAL =  LINEAR_MEM_SIZE / XVIRTUAL;

				mode->magnify.x = 1;
				mode->magnify.y = 1;

				/* QuickHack!! use roundmode?? */
				if (XVIRTUAL < XVISIBLE)
					XVIRTUAL=XVISIBLE;
				if (YVIRTUAL < YVISIBLE)
					YVIRTUAL=YVISIBLE;
				
			/* Up to 2x magnify is possible horizontally, and more vertically.
			 * Only 2x is supported at the moment--it's not really necessary,
			 * but results in a higher pixel clock.  An example mode is 
			 * 320x200: it uses 25 MHz (like VGA) and the same parameters as
			 * 640x400, so switching from standard text mode is instantaneous
			 * and mode centering is the same.  This may be obsoleted by
			 * the new monitor code. */
		    
			/* Tseng was smart and let this chip double pixels in
			 * accelerated mode, unlike S3--their Trio chips can't
			 * really handle 320x200 at less than 90Hz vertical
			 * refresh. */

				if ((XVISIBLE * mode->magnify.x) < 400)
					mode->magnify.x++;

			/* change to while for more magnification */
				if ((YVISIBLE * mode->magnify.y) < 350) 
					mode->magnify.y++;
			
				if ((XVIRTUAL * YVIRTUAL * bpp / 8) > LINEAR_MEM_SIZE) {
					if ((XVISIBLE * YVISIBLE * bpp / 8) > LINEAR_MEM_SIZE) {
						DEBUG1("Not enough memory to do virtual %ix%i at %i bpp",
							   XVIRTUAL, YVIRTUAL, bpp);
						return -ENOMEM;
					} 
					XVIRTUAL = XVISIBLE;
					YVIRTUAL = YVISIBLE;
				}
				mode->request.visible.x &= 0x7FFFFFF8;
				mode->request.visible.y &= 0x7FFFFFF8;
				mode->x.width = mode->request.visible.x * mode->magnify.x;
				mode->y.width = mode->request.visible.y * mode->magnify.y;
			}

			return EOK;
		
		case CMD_LOWER:

			if (mode->dclk < chipinfo.dclk.min) {
				DEBUG1("DCLK too low (%i)", mode->dclk);
				return -E(CHIPSET,FAILED);
			}
			mode->dclk = (mode->dclk > chipinfo.dclk.max)
				? chipinfo.dclk.max : mode->dclk;

			adjust_timing(mode);
			return EOK;

		case CMD_RAISE:

			if (mode->dclk > chipinfo.dclk.max) {
				DEBUG1("DCLK too high (%i)", mode->dclk);
				return -E(CHIPSET,FAILED);
			}
			mode->dclk = (mode->dclk < chipinfo.dclk.min)
				? chipinfo.dclk.min : mode->dclk;

			adjust_timing(mode);
			return EOK;

		case CMD_CHECK:
			if (mode->lclk.mul != 1) {
				DEBUG1("Invalid lclk multiplier %d", mode->lclk.mul);
				return -E(CHIPSET,FAILED);
			}
			switch (mode->lclk.div) {
				case 1:
					break;
				default:
					DEBUG1("Invalid CLK to LCLK divider %d", mode->lclk.div);
					return -E(CHIPSET,FAILED);
			}
			if (mode->request.graphtype == KGIGT_TEXT16) {
				mode->flags |= FL_CHIP_COLORTEXT;
				if ((mode->magnify.x != 1) || (mode->magnify.y != 1)
					|| (textx < 8) || (9 < textx) || (texty < 1)
					|| (texty > 32) || 
					((2 * XVIRTUAL * YVIRTUAL) > TEXT_LEN)
					|| (mode->dclk > chipinfo.dclk.max)
					|| (mode->dclk < chipinfo.dclk.min))
				{
					DEBUG1("HW check failed");
					return -E(CHIPSET,FAILED);
				}

			} else {

				/* There used to be some XVIRTUAL checking here. */

				if (((XVIRTUAL * YVIRTUAL * bpp / 8) > LINEAR_MEM_SIZE)
					|| (mode->magnify.x > 2) || (mode->magnify.y > 2)
					|| (mode->clk.mul != 1) || (mode->clk.div != 1)
					|| (mode->dclk < chipinfo.dclk.min)
					|| (chipinfo.dclk.max < mode->dclk)
					/* LOOKHERE -- allow double pels in > 8bpp modes -- not good code */
					/*|| (mode->magnify.x == 2 && mode->request.graphtype != KGIGT_8BIT) */ )
				{
					NOTICE1("HW check failed\n");
					return -E(CHIPSET,FAILED);
				}
			}

#define TM(X) ((chipinfo.maxdots.x < mode->x.X) || (chipinfo.maxdots.y < mode->y.X))
			if (TM(width) || TM(blankstart) || TM(syncstart)
				|| TM(syncend) || TM(blankend) || TM(total)) {
				NOTICE1("Timing check failed");
				DEBUG1("Maxdots are (%d,%d)", chipinfo.maxdots.x,
					   chipinfo.maxdots.y);
				return -E(CHIPSET,FAILED);
			}
#undef TM

			chipset_setup_mode(dpy, mode);
	        adjust_timing(mode);
	    
			return EOK;

		default:
  	        INTERNAL_ERROR;
			return -E(CHIPSET,FAILED);
	}
	
#undef XVIRTUAL
#undef YVIRTUAL
#undef XVISIBLE
#undef YVISIBLE
}

static void chipset_set_mode(struct kgi_display *dpy,struct kgi_mode *mode)
{
	struct statetable setup = default_state;
	register int foo;
	int DCLKdiv;
	int textx,texty;

	textx = mode->request.dpp.x;
	texty = mode->request.dpp.y;

	switch (mode->request.graphtype) {
		case KGIGT_TEXT16:
			foo = mode->request.virt.x >> 1; /* virt * 2 / 4 */
			break;
		case KGIGT_1BIT:
		case KGIGT_4BIT:
		    foo = mode->request.virt.x >> 4;
			break;
		case KGIGT_8BIT:
		    foo = mode->request.virt.x >> 3;
			break;
		case KGIGT_15BIT:
		case KGIGT_16BIT:
		    foo = mode->request.virt.x >> 2;
			break;
		case KGIGT_24BIT:
		    foo = (mode->request.virt.x >> 3) * 3;
			break;
		case KGIGT_32BIT:
		    foo = mode->request.virt.x >> 1;
		    break;
		default:
			INTERNAL_ERROR;
			return;
	}

	setup.CRT[CRT_LOGICALWIDTH] |= foo & 0xFF;
       	setup.CRT[CRT_OVRFLW_H] |= (foo & 0x100) ? CR3F_LOGICALWIDTH_B8 : 0;
	setup.CRT[CRT_OVRFLW_H] |= (foo & 0x200) ? CR3F_LOGICALWIDTH_B9 : 0;
    
	if (mode->request.graphtype == KGIGT_TEXT16) {
		setup.CRT[CRT_MODE] = CR17_CGA_ADDR_WRAP | CR17_ENABLE_SYNC | 
			CR17_CGA_BANKING | CR17_HGC_BANKING;

		setup.CRT[CRT_UNDERLINE] |= CR14_UNDERLINE_MASK;

		KGI_POKEBITS((SR02_PLANE0 | SR02_PLANE1),
			     SR02_PLANE_MASK, setup.SEQ[SEQ_WPLANE]);

		setup.GRC[GRC_MODECONTROL] |= GR5_ODD_EVEN;
		setup.GRC[GRC_MMAP_MODE] |= GR6_CHAIN_OE;

		KGI_POKEBITS(GR6_MAP_B8_32, GR6_MAP_MASK,
		             setup.GRC[GRC_MMAP_MODE]);

		switch ((DCLKdiv = textx)) {
			case 8:
				setup.SEQ[SEQ_CLOCK] |= SR01_8DOT_CHCLK;
				setup.ATC[ATC_MODECONTROL] &= ~AR10_9DOT_CHARS;
				break;
			case 9:
				setup.SEQ[SEQ_CLOCK] &= ~SR01_8DOT_CHCLK;
				KGI_POKEBITS(AR10_9DOT_CHARS,
				            (AR10_9DOT_CHARS | AR10_TOP_PAN),
				             setup.ATC[ATC_MODECONTROL]);
				break;
		 	default:
				INTERNAL_ERROR;
				return;
		}

		setup.CRT[0x09] |= texty - 1;
		setup.ATC[0x13] = (textx == 9) ? 0x08 : 0x00;
	} else {	/* mode->request.graphtype == KGIGT_TEXT16 */
		DCLKdiv = 8;

		setup.CRT[0x17] = CR17_CGA_BANKING | CR17_HGC_BANKING
			| CR17_ENABLE_SYNC | CR17_CGA_ADDR_WRAP;

		for (foo = 0; foo < 16; foo++)
			setup.ATC[foo] = foo;

		setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
		setup.SEQ[0x02] |= SR02_PLANE_MASK;
		setup.SEQ[0x04] |= SR04_256K_ACCESS | SR04_NO_ODDEVEN;

		setup.ATC[0x10] |= AR10_GRAPHICS | AR10_TOP_PAN;

		setup.GRC[0x06] |= GR6_GRAPHMODE | GR6_MAP_A0_64;

		setup.CRT[0x03] |= CR03_IS_VGA;
		setup.CRT[0x09] |= (mode->magnify.y - 1) & CR09_CHARHEIGHT_MASK;

		switch (mode->request.graphtype) {
		    case KGIGT_1BIT:
			case KGIGT_4BIT:
				if (mode->request.graphtype == KGIGT_1BIT) {
					setup.SEQ[0x02] = SR02_PLANE_MASK;
					setup.GRC[0x01] = 0x0e;	/* Set/Reset all but plane 0 */
					setup.GRC[0x04] = 0;
					setup.GRC[0x05] |= GR5_WRITEMODE0;
				} else {
					setup.GRC[0x01] = 0x0f; /* Set/Reset all planes */
					setup.GRC[0x05] |= GR5_WRITEMODE3;
				}

				setup.CRT[0x17] |= CR17_BYTE_MODE;
				break;
			case KGIGT_8BIT:
			case KGIGT_15BIT:
			case KGIGT_16BIT:
			case KGIGT_24BIT:
			case KGIGT_32BIT:
		    
				setup.SEQ[0x02] |= 0x0f;
				setup.SEQ[0x04] |= SR04_CHAINED;
				setup.CRT[0x14] |= CR14_DOUBLEWORDMODE;
		    
				/* Double the pixels for low-res modes, just like VGA.  This
				 * is not strictly necessary as the ET6000 clocks can handle
				 * frequencies of 12.75 Mhz (320x200), whereas the VGA uses 25 Mhz
				 * and sets this ATC bit. It may be better to remove this, unless
				 * low dot clocks make you uncomfortable.  Experimental.
				 */
									      
				if (mode->magnify.x == 2) {
					setup.ATC[0x10] |= AR10_8BIT_MODE;
				} else {
					setup.CRT[0x17] |= CR17_WORDMODE; /* inc memory addr only every 2 cycles */
					setup.CRT[0x14] |= CR14_MEMCOUNT_BY_4;
				}
		    
				setup.ATC[0x11] =  0x00;    /* overscan */
				setup.GRC[0x05] |= GR5_8BIT_COLOR;
				setup.GRC[0x08] |= 0xff;	/* Enable all bits */
		     
				break;
			default:
				INTERNAL_ERROR;
				return;
		}
		/* note !! fix this, by introducing bpp variable */
	        /* It's unlikely, but PCIVidCtrl may already contain bits which could
		 * be overwritten with zeros.  I will investigate.  */
		switch (mode->request.graphtype) {
			case KGIGT_15BIT:
				setup.PCIVidCtrl1 &= ~PCI58_BPP16;
				setup.ATC[0x16] |= (1 << 4) | 0x80;
				break;
			case KGIGT_16BIT:
				setup.ATC[0x16] |= (1 << 4) | 0x80;
				setup.PCIVidCtrl1 |= PCI58_BPP16;
				break;
			case KGIGT_24BIT:
				setup.ATC[0x16] |= (2 << 4) | 0x80;
				break;
			case KGIGT_32BIT:
				setup.ATC[0x16] |= (3 << 4) | 0x80;
				break;
			default:
		}
	}

	/* Set up horizontal timing */
	foo = (mode->x.width / DCLKdiv) - 1;	/* Horizontal display end */
	setup.CRT[0x01] |= foo & 0xFF;
   
	foo = (mode->x.blankstart / DCLKdiv);	/* Horizontal blank start */
	setup.CRT[0x02] |= foo & 0xFF;
    	setup.CRT[0x3F] |= (foo & 0x100) ? CR3F_HBLANKSTART_B8 : 0;

    
	foo = (mode->x.syncstart / DCLKdiv);	/* Horizontal sync start  */
	setup.CRT[0x04] |= foo & 0xFF;
	setup.CRT[0x3F] |= (foo & 0x100) ? CR3F_HSYNCSTART_B8 : 0;
    
	foo = (mode->x.syncend / DCLKdiv);	/* Horizontal sync end    */
	setup.CRT[0x05] |= foo & CR05_HSE_MASK;
     
	foo = (mode->x.blankend / DCLKdiv);	/* Horizontal blank end   */
	setup.CRT[0x03] |= foo & CR03_HBE_MASK;
	setup.CRT[0x05] |= (foo & 0x20) ? CR05_HBLANKEND_B5 : 0;
    
	/* Think about adding a little to this if it's the card and not the monitor. */
	/* The problem may be my card (Jaton) and not the ET6000. */
	foo = (mode->x.total / DCLKdiv) - 5;	/* Horizontal total       */
	setup.CRT[0x00] |= foo & 0xFF;
	setup.CRT[0x3F] |= (foo & 0x100) ? CR3F_HTOTAL_B8 : 0;

	/* Set up vertical timing */

	foo = mode->y.width - 1;		/* Vertical display end   */
	setup.CRT[0x12] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VDISPLAYEND_B8 : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VDISPLAYEND_B9 : 0;
	setup.CRT[0x35] |= (foo & 0x400) ? CR35_VDISPLAYEND_B10 : 0;
    
	foo = mode->y.blankstart - 1;		/* Vertical blank start   */
	setup.CRT[0x15] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VBLANKSTART_B8 : 0;
	setup.CRT[0x09] |= (foo & 0x200) ? CR09_VBLANKSTART_B9 : 0;
	setup.CRT[0x35] |= (foo & 0x400) ? CR35_VBLANKSTART_B10 : 0;

	foo = mode->y.syncstart;		/* Vertical sync start    */
	setup.CRT[0x10] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VSYNCSTART_B8 : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VSYNCSTART_B9 : 0;
	setup.CRT[0x35] |= (foo & 0x400) ? CR35_VSYNCSTART_B10 : 0;

	foo = mode->y.syncend;			/* Vertical sync end      */
	setup.CRT[0x11] |= foo & CR11_VSYNCEND_MASK;

	foo = mode->y.blankend + 1;		/* Vertical blank end     */
	setup.CRT[0x16] |= foo & 0xFF;

	foo = mode->y.total - 2;		/* Vertical total         */
	setup.CRT[0x06] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VTOTAL_B8 : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VTOTAL_B9 : 0;
	setup.CRT[0x35] |= (foo & 0x400) ? CR35_VTOTAL_B10 : 0;
    
	setup.MISC |= (mode->x.polarity > 0) ? 0 : MISC_NEG_HSYNC;
	setup.MISC |= (mode->y.polarity > 0) ? 0 : MISC_NEG_VSYNC;

	/* Set splitline to 0 */

	foo = 0;
	setup.CRT[0x18] |= foo & 0xFF;

	setup.CRT[0x07] |= (foo & 0x0100) ? CR07_LINECOMPARE_B8 : 0;
	setup.CRT[0x09] |= (foo & 0x0200) ? CR09_LINECOMPARE_B9 : 0;
	setup.CRT[0x35] |= (foo & 0x0400) ? CR35_LINECOMPARE_B10 : 0;
    
	/* Ok.  We'll do the actual hardware programming now. */

	kgim_monitor_set_mode(dpy, mode);
	kgim_ramdac_set_mode(dpy, mode);
	kgim_clock_set_mode(dpy, mode);

	restore_state(&setup);
	     
	chipset_setup_dpy(dpy, mode);

	DEBUG2("Mode set complete\n");
}

static int chipset_command(struct kgi_graphic *state,uint cmd,uint arg)
{
	kgi_coord  *coord;
	struct kgi_font *font;
	struct kgi_display *dpy;

	dpy = state->dev.dpy;
	switch (cmd) {
		case CHIP_SETDEFFONT:
			font = (struct kgi_font *)(state->io_buf);
	        text16_set_font(dpy, font, 0, font->info->positions);
			return 0;
		case CHIP_SETSPLITLINE:
			if (arg > state->dev.mode.visible.y)
				return -E(CHIPSET,BOGUS);

			if (state->dev.org[0].y > state->dev.mode.visible.y - arg )
				kgi_set_origin(&(state->dev), 0, 
							   state->dev.org[0].x,
							   state->dev.mode.visible.y - arg );
		
			kgi_set_split(&(state->dev),arg*dpy->mode.magnify.y);
			return 0;
		case CHIP_SETVISFRAME:
			coord = (kgi_coord *)(state->io_buf);

			if (coord->x < 0 || 
			    coord->x > state->dev.mode.virt.x - state->dev.mode.visible.x)
				return -E(CHIPSET,BOGUS);

			if (coord->y < 0 || dpy->mode.magnify.y<1 ||
			    coord->y > state->dev.mode.virt.y - state->dev.split/dpy->mode.magnify.y)
				return -E(CHIPSET,BOGUS);

			kgi_set_origin(&(state->dev), 0, coord->x, coord->y);
			return 0;
		default:
			break;
	}
	return -E(CHIPSET,UNKNOWN);
}

kgi_sint kgim_chipset_init(struct kgi_card *card,struct kgi_display *dpy)
{
	int err;
	struct chipset_private *priv;
    
	priv = kgi_malloc_small(sizeof(*priv));
	if (priv == NULL) {
		NOTICE2("ERROR: failed to allocate private data!");
		return -ENOMEM;
	}
	memset(priv, 0, sizeof(*priv));
	KGIM_PRIVATE(dpy, chipset) = priv;   
	DEBUG2("Private data at 0x%p (0x%p)\n", priv, KGIM_PRIVATE(dpy, chipset));
	if ((err = chipset_init(dpy)) != EOK) return err;
 
	strcpy(card->manufact, chipinfo.manufact);
	strcpy(card->model, chipinfo.model);
	card->modes = chipinfo.modes;
	card->flags = chipinfo.flags;
	card->maxdots = chipinfo.maxdots;

	card->chipset =& chipinfo;
	card->preset = default_kgi_mode;

	return EOK;
}

void kgim_chipset_done(struct kgi_display *dpy)
{
	chipset_done();
}

int  kgim_chipset_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
			     enum kgi_tm_cmd *cmd)
{
	int val;
	DEBUG1("Checking mode... (%dx%d %d)\n", mode->request.virt.x,
	       mode->request.virt.y, mode->request.graphtype);
	val = chipset_check_mode(dpy, mode, *cmd);
	DEBUG2("Mode check %s\n", (val == EOK) ? "succeeded" : "failed");
    return val;
}

void kgim_chipset_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	DEBUG1("Setting mode... (%dx%d %d)\n", mode->request.virt.x,
	mode->request.virt.y, mode->request.graphtype);
	chipset_set_mode(dpy,mode);
}

kgi_sint kgim_chipset_command(struct kgi_graphic *state, uint cmd, uint iobuf)
{
	return chipset_command(state, cmd, iobuf);
}
